from pathlib import Path
from typing import Dict, List

import yaml
from pydantic import BaseModel, validator

from load_atoms.checksums import valid_checksum
from load_atoms.util import BASE_REMOTE_URL, DATASETS_DIR


def is_bibtex(data: str) -> bool:
    data = data.strip()
    return data.startswith("@") and data.endswith("}")


class DatasetDescription(BaseModel):
    name: str
    description: str
    files: Dict[str, str]
    citation: str = None
    license: str = None
    representative_structures: List[int] = None
    long_description: str = None
    per_atom_properties: dict = None
    per_structure_properties: dict = None
    url_root: str = BASE_REMOTE_URL

    @validator("license")
    def validate_license(cls, v):
        valid_licences = ["MIT", "CC-BY-4.0", "CC BY-NC-SA 4.0"]
        if v not in valid_licences:
            raise ValueError(f"Invalid license: {v}. Must be one of {valid_licences}")
        return v

    @validator("files")
    def validate_files(cls, v):
        for data in v.values():
            if not valid_checksum(data):
                raise ValueError(f"Invalid checksum: {data}")
        return v

    @validator("citation")
    def validate_citation(cls, v):
        if not is_bibtex(v):
            raise ValueError(f"Invalid BibTeX: {v}")
        return v

    @classmethod
    def from_file(cls, path: Path) -> "DatasetDescription":
        with open(path) as f:
            data = yaml.safe_load(f)
        return cls(**data)


def find_all_descriptor_files() -> List[Path]:
    """Find all descriptor files."""

    return list(DATASETS_DIR.glob("**/*.yaml"))


def is_known_dataset(dataset_id: str) -> bool:
    """Check if a dataset is known."""
    return dataset_id in DATASETS


def get_description_of(dataset_id: str) -> DatasetDescription:
    """Get the database entry for a dataset."""
    assert is_known_dataset(dataset_id), f"Dataset {dataset_id} is not known."
    return DATASETS[dataset_id]


DATASETS = {}
for file in find_all_descriptor_files():
    entry = DatasetDescription.from_file(file)
    DATASETS[entry.name] = entry



from typing import Callable, Sequence, Union

import numpy as np
from ase import Atoms

from load_atoms.dataset import Dataset

FilterFunction = Callable[[Atoms], bool]


def filter_by(
    dataset: Union[Dataset, Sequence[Atoms]],
    *functions: Sequence[FilterFunction],
    **kwargs,
) -> Dataset:
    """Filter a dataset by a given property."""

    structures_filtered_by_kwargs = [
        structure
        for structure in dataset
        if all(structure.info.get(key, None) == value for key, value in kwargs.items())
    ]

    return Dataset(
        [
            structure
            for structure in structures_filtered_by_kwargs
            if all(func(structure) for func in functions)
        ]
    )


def cross_validate_split(
    dataset: Dataset, fold, k=5, n_test: int = None, seed: int = 0
) -> Dataset:
    """Generate a shuffled train/test split for cross-validation."""

    if n_test is None:
        n_test = len(dataset) // k

    idxs = np.arange(len(dataset))
    np.random.RandomState(seed).shuffle(idxs)
    idxs = np.roll(idxs, fold * len(idxs) // k)
    train, test = idxs[:-n_test], idxs[-n_test:]

    return (
        Dataset.from_structures([dataset[t] for t in train]),
        Dataset.from_structures([dataset[t] for t in test]),
    )



from pathlib import Path

import pytest
from pydantic import ValidationError

from load_atoms.checksums import generate_checksum
from load_atoms.database import (
    DATASETS,
    DatasetDescription,
    find_all_descriptor_files,
    get_description_of,
)
from load_atoms.util import BASE_REMOTE_URL

_local_path_to_datasets = Path(__file__).parent.parent / "src/load_atoms/datasets"


def test_all_loaded():
    """
    test that all datasets are loaded
    """

    assert len(DATASETS) == len(find_all_descriptor_files())


@pytest.mark.parametrize("dataset", DATASETS.values(), ids=lambda x: x.name)
def test_checksums(dataset):
    """
    test that all dataset checksums are correct
    """

    # we don't want to download all datasets each time we run the tests
    if dataset.url_root != BASE_REMOTE_URL:
        return

    for path, checksum in dataset.files.items():
        path = _local_path_to_datasets / path
        assert checksum == generate_checksum(path)


def test_description():
    desc = get_description_of("C-GAP-17")
    assert desc.name == "C-GAP-17"


def get_correct_dict():
    return DATASETS["C-GAP-17"].dict()


def test_incorrect_name():
    kwargs = get_correct_dict()
    kwargs["name"] = dict(_not="a string")

    with pytest.raises(ValidationError):
        DatasetDescription(**kwargs)


def test_incorrect_license():
    kwargs = get_correct_dict()
    kwargs["license"] = "Not a real license"

    with pytest.raises(ValidationError):
        DatasetDescription(**kwargs)


def test_incorrect_files():
    kwargs = get_correct_dict()

    # files should be a mapping of str to checksum

    # not a mapping
    kwargs["files"] = 1
    with pytest.raises(ValidationError):
        DatasetDescription(**kwargs)

    # not a mapping of str to checksum
    kwargs["files"] = {1: 1}
    with pytest.raises(ValidationError):
        DatasetDescription(**kwargs)


def test_incorrect_citation():
    kwargs = get_correct_dict()

    # citation should be a bibtex string
    kwargs["citation"] = "this is not a bibtex string"
    with pytest.raises(ValidationError):
        DatasetDescription(**kwargs)



from pathlib import Path
from typing import Iterable, List

from ase import Atoms
from ase.io import read
from yaml import dump

from load_atoms import backend
from load_atoms.database import DATASETS, DatasetDescription, is_known_dataset
from load_atoms.util import frontend_url, intersection, union


class Dataset:
    def __init__(
        self, structures: Iterable[Atoms], description: DatasetDescription = None
    ):
        self.structures = [*structures]
        self._description = description

    def __len__(self):
        return len(self.structures)

    def __getitem__(self, slice):
        # if the slice can be cast to an int (e.g. is an np.int64), then
        # return a single structure
        try:
            slice = int(slice)
            return self.structures[slice]
        except TypeError:
            pass

        # otherwise, our slice is a slice object,
        # and we should return a new Dataset
        sliced_structres = self.structures[slice]
        return Dataset(sliced_structres)

    def __iter__(self):
        return iter(self.structures)

    def __repr__(self):
        return summarise_dataset(self.structures, self._description)

    @classmethod
    def from_structures(cls, structures: Iterable[Atoms]):
        return cls(structures)

    @classmethod
    def from_file(cls, path: Path):
        structures = read(path, index=":")
        return cls(structures, path.stem)

    @classmethod
    def from_id(cls, dataset_id: str, root: Path = None, verbose: bool = True):
        if not is_known_dataset(dataset_id):
            raise ValueError(f"Dataset {dataset_id} is not known.")

        dataset_info = DATASETS[dataset_id]
        all_structures = backend.get_structures(dataset_info, root)
        if verbose:
            print(usage_info(dataset_info))
        return cls(all_structures, dataset_info)

    @classmethod
    def from_structures(cls, structures: Iterable[Atoms]):
        return cls(structures)

    @classmethod
    def from_file(cls, path: Path):
        structures = read(path, index=":")
        return cls(structures)


def usage_info(dataset: DatasetDescription) -> str:
    info = []
    if dataset.license is not None:
        info.append(f"This dataset is covered by the {dataset.license} license.")
    if dataset.citation is not None:
        info.append(f"Please cite this dataset if you use it in your work.")

    _url = frontend_url(dataset)
    info.append(f"For more information, visit:\n{_url}")

    return "\n".join(info)


def summarise_dataset(
    structures: List[Atoms], description: DatasetDescription = None
) -> str:
    name = description.name if description is not None else "Dataset"
    N = len(structures)
    number_atoms = sum([len(structure) for structure in structures])

    per_atom_properties = intersection(
        structure.arrays.keys() for structure in structures
    )
    per_atom_properties -= {"numbers", "positions"}
    per_structure_properties = intersection(
        structure.info.keys() for structure in structures
    )

    species = union(structure.get_chemical_symbols() for structure in structures)
    species_counts = {
        s: sum([structure.get_chemical_symbols().count(s) for structure in structures])
        for s in species
    }
    species_counts = {
        k: f"{v / number_atoms:.2%}"
        for k, v in sorted(
            species_counts.items(), key=lambda item: item[1], reverse=True
        )
    }

    if N >= 1000:
        N = f"{N:,}"
    if number_atoms >= 1000:
        number_atoms = f"{number_atoms:,}"

    fields = {
        "structures": N,
        "atoms": number_atoms,
        "species": species_counts,
        "properties": {
            "per atom": "(" + ", ".join(per_atom_properties) + ")",
            "per structure": "(" + ", ".join(per_structure_properties) + ")",
        },
    }

    return dump({name: fields}, sort_keys=False, indent=4)



"""
The backend is responsible for downloading the datasets when they 
are first loaded, and for loading these datasets into memory, via
the `get_structures` function. 
"""

import asyncio
from pathlib import Path
from typing import List

import aiohttp
import nest_asyncio
from ase import Atoms
from ase.io import read
from rich.progress import Progress, track

from load_atoms.checksums import generate_checksum
from load_atoms.database import DatasetDescription
from load_atoms.util import DEFAULT_DOWNLOAD_DIR

# support for nested asyncio loops, such that we can download
# structures inside a jupyter notebook in parallel
nest_asyncio.apply()


def get_event_loop():
    loop = asyncio.get_event_loop()
    if loop.is_running():
        loop = asyncio.get_running_loop()
    else:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


class RequestError(Exception):
    def __init__(self, url: str, code: int):
        message = f"Could not find {url}. Response code: {code}"
        super().__init__(message)


class CorruptionError(Exception):
    def __init__(self, path: Path):
        super().__init__(
            f"File {path} exists, but did not match the expected checksum."
            "Please move this file to a different location, or delete it."
        )


def get_structures(dataset: DatasetDescription, root: Path = None) -> List[Atoms]:
    """
    Get the structures associated with a dataset.
    """

    # if no root is specified, use the default download directory
    if root is None:
        root = DEFAULT_DOWNLOAD_DIR
    root = Path(root)

    # first, we download any missing files:
    loop = get_event_loop()
    loop.run_until_complete(download_missing_files(dataset, root))

    # now, we can load the structures from the files
    all_structures = []
    for file, hash in track(
        dataset.files.items(), description="Loading files from disk"
    ):
        local_path = root / file
        check_file_contents(local_path, hash)
        structures = read(local_path, index=":")
        all_structures.extend(structures)

    return all_structures


async def download_missing_files(
    dataset: DatasetDescription, root: Path = None
) -> None:
    missing_files = [file for file in dataset.files if not (root / file).exists()]
    if not missing_files:
        return

    tasks = []
    with Progress(transient=True) as progress:
        progress.console.print(f"Downloading {len(missing_files)} files.")
        for file in missing_files:
            local_path = root / file
            remote_url = dataset.url_root + file

            download_task = download_structures(
                remote_url,
                local_path,
                progress=progress,
            )
            tasks.append(download_task)

        # wait for all downloads to finish
        await asyncio.wait(tasks)


def check_file_contents(local_path: Path, expected_file_hash: str) -> bool:
    file_hash = generate_checksum(local_path)
    if expected_file_hash != file_hash:
        raise CorruptionError(
            f"The dataset on disk at {local_path} has been corrupted.\n"
            "If you have made changes to this, "
            "please revert them or move/delete the file."
        )
    return True


async def download_structures(
    remote_url: str,
    local_path: Path,
    progress: Progress,
):
    local_path.parent.mkdir(parents=True, exist_ok=True)
    await download_file(
        remote_url, local_path, progress, task_description=local_path.name
    )

    if not local_path.exists():
        raise ValueError(
            "There was a problem downloading the dataset.\nPlease try again."
        )


async def download_file(
    url: str, save_to: Path, progress: Progress = None, task_description: str = None
):
    """
    Download a file from the internet, optionall using a Rich Progress bar
    to show the progress.
    """

    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            if response.status != 200:
                raise RequestError(url, response.status)

            total_length = int(response.headers.get("content-length", 0))
            if progress is not None:
                task = progress.add_task(task_description, total=total_length)

            with save_to.open("wb") as f:
                async for chunk in response.content.iter_chunked(1024):
                    f.write(chunk)
                    if progress is not None:
                        progress.update(task, advance=len(chunk))

            if progress is not None:
                progress.remove_task(task)



from pathlib import Path
from typing import Iterable

# this file is at <root>/src/load_atoms/util.py
# the datasets are at <root>/src/load_atoms/datasets
DATASETS_DIR = (Path(__file__).parent / "datasets").resolve()
DEFAULT_DOWNLOAD_DIR = Path.home() / ".load_atoms"
BASE_REMOTE_URL = (
    "https://github.com/jla-gardner/load-atoms/raw/main/src/load_atoms/datasets/"
)
FRONTEND_URL = "https://jla-gardner.github.io/load-atoms/datasets/"


def frontend_url(dataset_info):
    """Get the URL for a dataset."""
    return FRONTEND_URL + dataset_info.name + ".html"


def lpad(s, indent=4):
    return "\n".join(" " * indent + line for line in s.splitlines())


def intersection(things: Iterable[Iterable]):
    """Get the set intersection of a list of iterables."""
    return set.intersection(*map(set, things))


def union(things: Iterable[Iterable]):
    """Get the set union of a list of iterables."""
    return set.union(*map(set, things))



from ase import Atoms

from load_atoms.manipulations import cross_validate_split, filter_by

structures = [
    Atoms("H2O", info=dict(name="water")),
    Atoms("H2O2", info=dict(name="hydrogen peroxide")),
    Atoms("CH4", info=dict(name="methane")),
]


def test_filter_by():
    filtered = filter_by(
        structures,
        lambda structure: len(structure) >= 4,
    )
    assert len(filtered) == 2, "The filtered dataset should contain two structures"

    filtered = filter_by(
        structures,
        name="water",
    )
    assert len(filtered) == 1, "The filtered dataset should contain one structure"


def test_cv():
    train, test = cross_validate_split(structures, fold=1, k=3)
    assert len(train) == 2, "The train dataset should contain two structures"
    assert len(test) == 1, "The test dataset should contain one structure"



from load_atoms.util import intersection, lpad, union


def test_intersection():
    assert intersection([[1, 2, 3], [2, 3, 4]]) == {2, 3}
    assert intersection(("hi", "hello")) == {"h"}


def test_union():
    assert union([[1, 2, 3], [2, 3, 4]]) == {1, 2, 3, 4}
    assert union(("hi", "hello")) == {"h", "e", "l", "o", "i"}


def test_pad():
    assert lpad("hi", 2) == " " * 2 + "hi"
    assert lpad("hi\nthere", 2) == "  hi\n  there"



import hashlib
import string
from pathlib import Path


def generate_checksum(file_path: Path) -> str:
    """Generate a checksum for a file."""

    sha256_hash = hashlib.sha256()
    with open(file_path, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)

    return sha256_hash.hexdigest()[:12]


def valid_checksum(hash: str) -> bool:
    """Check if a hash is valid."""
    if len(hash) != 12:
        return False
    return all(c in string.hexdigits for c in hash)



from pathlib import Path
from typing import List, Union

from ase import Atoms

from load_atoms.dataset import Dataset


def dataset(
    thing: Union[str, List[Atoms], Path],
    root: Union[str, Path] = None,
) -> Dataset:
    """
    Load a dataset by name or from a list of structures.

    Parameters
    ----------
    thing
        A dataset id, a list of structures, or a path to a file.
    root
        The root directory to use when loading a dataset by id.
        If not provided, the default root directory will be used.

    Returns
    -------
    Dataset
        The loaded dataset.

    Examples
    --------
    >>> from load_atoms import dataset
    >>> from ase import Atoms
    >>> from ase.io import read, write
    >>> dataset("qm7")
    >>> dataset("qm7", root="./my-datasets")
    >>> dataset([Atoms("H2O"), Atoms("H2O2")])
    >>> dataset("path/to/file.xyz")
    """

    if isinstance(thing, list) and all(isinstance(s, Atoms) for s in thing):
        # thing is a list of structures
        return Dataset.from_structures(thing)

    if not isinstance(thing, (Path, str)):
        raise TypeError(
            f"Could not load dataset from {thing}. "
            "Please provide a string, a list of structures, or a path to a file."
        )

    if Path(thing).exists():
        # thing is a string/path to a file that exists
        # assume it is a file containing structures and load them
        return Dataset.from_file(Path(thing))

    if isinstance(thing, Path):
        # thing is a path to a file that does not exist
        raise ValueError(f"The provided path does not exist. ({thing})")

    # assume thing is a dataset ID, and try to load it
    return Dataset.from_id(thing, root)



from pathlib import Path

import pytest
from ase import Atoms
from ase.io import read, write

from load_atoms import dataset
from load_atoms.dataset import Dataset, summarise_dataset

STRUCTURES = [Atoms("H2O"), Atoms("H2O2")]


def _is_water_dataset(ds):
    assert len(ds) == 2
    assert set(ds[0].symbols) == {"H", "O"}


def test_dataset_from_structures():
    ds = dataset(STRUCTURES)
    _is_water_dataset(ds)


def test_dataset_writeable_and_readable(tmp_path):
    ds = dataset(STRUCTURES)
    write(tmp_path / "test.xyz", ds)

    ds2 = dataset(read(tmp_path / "test.xyz", index=":"))
    _is_water_dataset(ds2)

    ds3 = dataset(tmp_path / "test.xyz")
    _is_water_dataset(ds3)


def test_slice():
    ds = dataset(STRUCTURES)

    structure = ds[0]
    assert isinstance(structure, Atoms), "Indexing should return an Atoms object"
    assert structure is STRUCTURES[0], "Indexing should return the same structure"

    sub_dataset = ds[1:]
    assert isinstance(sub_dataset, Dataset), "Slicing should return a dataset"
    assert (
        len(sub_dataset) == 1
    ), "Slicing should return the correct number of structures"


def test_can_load_from_id():
    # pass root to avoid downloading the dataset
    structures = dataset("QM7", root="src/load_atoms/datasets")
    assert len(structures) == 7165

    with pytest.raises(ValueError, match="is not known."):
        dataset("made_up_dataset")


def test_summarise():
    ds = dataset(STRUCTURES)
    summary = summarise_dataset(ds)
    assert "Dataset" in summary, "The summary should contain the dataset name"

    ds = dataset("C-GAP-17", root="src/load_atoms/datasets")
    summary = summarise_dataset(ds)
    assert "energy" in summary, "The summary should contain the property names"


def test_useful_error_message():
    with pytest.raises(
        TypeError,
        match="Please provide a string, a list of structures, or a path to a file.",
    ):
        dataset(1)

    with pytest.raises(
        ValueError,
        match="The provided path does not exist.",
    ):
        dataset(Path("made_up_file.xyz"))



import pytest
import requests
from aiohttp import ClientConnectorError

from load_atoms.backend import RequestError, download_file, download_structures
from load_atoms.util import DATASETS_DIR


@pytest.mark.asyncio
async def test_real_download(tmp_path):
    url = "https://raw.githubusercontent.com/jla-gardner/load-atoms/main/README.md"

    # download the file
    path = tmp_path / "test.json"
    await download_file(url, path)

    # check that the file exists
    assert path.exists(), "The file was not downloaded"


@pytest.mark.asyncio
async def test_missing_download(tmp_path):
    # attempt to download a file from a non-existent url
    fake_url = "https://this.is.a.fake.url/file.json"
    with pytest.raises(ClientConnectorError):
        await download_file(fake_url, tmp_path / "test.json")

    # attempt to download a file that does not exist
    # and check that we report on the 404 error
    real_url_fake_file = "https://raw.githubusercontent.com/jla-gardner/load-atoms/main/made_up_file.json"
    with pytest.raises(RequestError, match="404"):
        await download_file(real_url_fake_file, tmp_path / "test.json")


@pytest.mark.asyncio
async def test_download_structures(tmp_path):
    # test that download structures raises an error if an existing
    # file is corrupted

    fake_url = "https://this.is.a.fake.url/file.extxyz"
    fake_file = tmp_path / "file.extxyz"
    fake_file.write_text("fake file")
    incorrect_hash = "0" * 12

    with pytest.raises(ClientConnectorError):
        await download_structures(fake_url, fake_file, incorrect_hash)

    existing = DATASETS_DIR / "QM7.extxyz"
    new = tmp_path / "QM7.extxyz"
    new.write_bytes(existing.read_bytes())
